#include "common/common.hpp"
#include "packet.hpp"
#include "common/utilities.hpp"

using boost::asio::buffer;
using boost::asio::read;
using boost::asio::streambuf;
using boost::asio::transfer_exactly;
using boost::asio::write;

packet::packet ( ReceiveCallBackType pReceiveFn ) noexcept
: m_pReceiveFunction ( pReceiveFn )
{
    lock_stream ( );
    LOG_ERR ( ) << "Constructor" << std::endl;
    unlock_stream ( );
    m_dataBuffer.clear ( );
}

packet::~packet ( ) noexcept
{
    lock_stream ( );
    LOG_ERR ( ) << "Destructor" << std::endl;
    unlock_stream ( );
}

error_code packet::recv ( tcp::socket &socket ) noexcept
{
    std::string threadId = boost::lexical_cast<std::string>( boost::this_thread::get_id ( ) );
    lock_stream ( );
    LOG_ERR ( ) << "ThreadId : " << threadId << std::endl;
    LOG_INF ( ) << " [" << get_peer_ip ( socket ) << " : " << get_peer_port ( socket ) << "]" << std::endl;
    unlock_stream ( );
    error_code ec;
    try
    {
        streambuf read_buffer;
        unsigned int data_read = read ( socket, read_buffer , ec );
        lock_stream ( );
        LOG_ERR ( ) << "data_read " << data_read << std::endl;
        unlock_stream ( );		
        const std::string data_received ( ( std::istreambuf_iterator<char>( &read_buffer ) ),
                                          std::istreambuf_iterator<char>( ) );


        m_dataBuffer.clear ( );
        std::copy ( data_received.begin ( ), data_received.end ( ), std::back_inserter ( m_dataBuffer ) );

        m_pReceiveFunction ( data_received, get_peer_port ( socket ) );
        ec.assign ( 0, boost::system::system_category ( ) );

    }
    catch ( const std::exception &e )
    {
        lock_stream ( );
        LOG_ERR ( ) << "Exception: " << e.what ( ) << std::endl;
        unlock_stream ( );
    }

    return ec;
}

error_code packet::send ( tcp::socket &socket ) noexcept
{
    std::string threadId = boost::lexical_cast<std::string>( boost::this_thread::get_id ( ) );
    lock_stream ( );
    LOG_ERR ( ) << "ThreadId : " << threadId << std::endl;
    LOG_ERR ( ) << "Socket is open : " << socket.is_open ( ) << std::endl;
    unlock_stream ( );
    error_code ec ;
    if ( !socket.is_open ( ) || ( m_dataBuffer.size ( ) <= 0 ) )
    {
        lock_stream ( );
        LOG_INF ( ) << "Socket is not open or input buffer is empty " << std::endl;
        unlock_stream ( );
        ec.assign ( 2, boost::system::system_category ( ) );
        return ec;
    }
    try
    {
        lock_stream ( );
        LOG_INF ( ) << " [" << get_peer_ip ( socket ) << " : " << get_peer_port ( socket ) << "]" << std::endl;
        unlock_stream ( );
        // Send packet message
        write ( socket, buffer ( m_dataBuffer ), transfer_exactly ( m_dataBuffer.size ( ) ), ec );
        if ( ec )
        {
            lock_stream ( );
            LOG_ERR ( ) << "Could not send message! ["
                << get_peer_ip ( socket ) << ":"
                << get_peer_port ( socket )
                << "]" << std::endl;
            unlock_stream ( );

            return ec;
        }

    }
    catch ( const std::exception &e )
    {
        lock_stream ( );
        LOG_ERR ( ) << "Exception: " << e.what ( ) << std::endl;
        unlock_stream ( );
    }

    lock_stream ( );
    LOG_INF ( ) << "Packet sent to ["
        << get_peer_ip ( socket ) << ":"
        << get_peer_port ( socket )
        << "]..." << std::endl;
    unlock_stream ( );

    return ec;
}

error_code packet::process ( tcp::socket &socket ) noexcept
{
    std::string threadId = boost::lexical_cast<std::string>( boost::this_thread::get_id ( ) );
    lock_stream ( );
    LOG_ERR ( ) << "ThreadId : " << threadId << std::endl;
    LOG_INF ( ) << " [" << get_peer_ip ( socket ) << " : " << get_peer_port ( socket ) << "]" << std::endl;
    unlock_stream ( );
    // Communicate with client using defined protocol
    error_code ec;
    try
    {
        do
        {
            ec = recv ( socket );
            if ( ec )
            {
                lock_stream ( );
                LOG_ERR ( ) << "recv error: " << ec.message ( ) << std::endl;
                unlock_stream ( );
            }
        }
        while ( !ec && socket.is_open ( ) );
    }
    catch ( const std::exception &e )
    {
        lock_stream ( );
        LOG_ERR ( ) << "Exception: " << e.what ( ) << std::endl;
        unlock_stream ( );
    }

    lock_stream ( );
    LOG_INF ( ) << "Session ended with client for " 
        << get_peer_ip ( socket ) << ":"
        << get_peer_port ( socket )
        << "]..." << std::endl;       
    unlock_stream ( );

    return ec;
}


